import unittest

from flask import Flask, g
from flask.ext.jsontools import FlaskJsonClient, DynamicJSONEncoder
from sqlalchemy.orm.exc import NoResultFound

from . import models
from .crud_view import ArticlesView


class CrudTest(unittest.TestCase):
    def setUp(self):
        # Connect, create tables
        engine, Session = models.init_database()
        models.drop_all(engine)
        models.create_all(engine)

        # Fill DB
        ssn = Session()
        ssn.begin()
        ssn.add_all(models.content_samples())
        ssn.commit()

        # Session
        self.Session = Session
        self.engine = engine
        self.db = Session()
        self.db.begin()

        # Flask
        self.app = app = Flask(__name__)
        app.debug = app.testing = True
        app.json_encoder = DynamicJSONEncoder
        app.test_client_class = FlaskJsonClient

        ArticlesView.route_as_view(app, 'articles', ('/article/', '/article/<int:id>'))

        @app.before_request
        def db():
            g.db = self.db

    def tearDown(self):
        self.db.close()  # Reset session

    def test_list(self):
        """ Test list() """

        # Simple list
        # maxitems:2, sort:id- should apply
        with self.app.test_client() as c:
            rv = c.get('/article/', json=None)
            self.assertEqual(rv['articles'], [
                # 2 items
                # sort: id-
                {'id': 30, 'uid': 3, 'title': '30', 'data': {'o': {'z': False}},},
                {'id': 21, 'uid': 2, 'title': '21', 'data': {'rating': 4, 'o': {'z': True}},}
            ])

        # Query list
        # Try to override sort, limit
        with self.app.test_client() as c:
            rv = c.get('/article/', json={
                'query': {
                    'limit': 3, # Cannnot exceed
                    'sort': ['id+'],  # Sort changed
                    'project': ['id', 'uid']
                }})
            self.assertEqual(rv['articles'], [
                # Still 2 items: cannot exceed maxitems
                # sort: id+ (overridden)
                # Projection worked
                {'id': 10, 'uid': 1},
                {'id': 11, 'uid': 1},
            ])

        # Query list, aggregate
        with self.app.test_client() as c:
            rv = c.get('/article/', json={
                'query': {
                    'filter': {
                        'id': {'$gte': '10'},
                    },
                    'aggregate': {
                        'n': {'$sum': 1},
                        'sum_ids': {'$sum': 'id'},
                        'max_rating': {'$max': 'data.rating'},
                        'avg_rating': {'$avg': 'data.rating'},
                    },
                    'sort': None,  # Unset initial sorting. Otherwise, PostgreSQL wants this column in GROUP BY
                }})
            self.assertEqual(rv['articles'], [
                {
                    'n': 6,
                    'sum_ids': 10+11+12+20+21+30,
                    'max_rating': 6.0,
                    'avg_rating': (5+5.5+6+4.5+4  +0)/5,
                }
            ])

        # Test count
        with self.app.test_client() as c:
            rv = c.get('/article/', json={
                'query': {
                    'count': 1
                }})
            self.assertEqual(rv['articles'], 6)  # `maxitems` shouldnt apply here

    def test_create(self):
        """ Test create() """

        # Create
        # 'ro' field should be set manually
        with self.app.test_client() as c:
            rv = c.post('/article/', json={
                'article': {
                    'id': 999, 'uid': 999,
                    'title': '999',
                    'data': {'wow': True}
                }
                })
            self.assertEqual(rv['article'], {
                'id': 1,  # Auto-set
                'uid': 3,  # Set manually
                'title': '999',
                'data': {'wow': True}
            })

    def test_get(self):
        """ Test get() """

        # Simple get
        with self.app.test_client() as c:
            rv = c.get('/article/30', json={
                'query': {
                    'project': ['uid'],
                }
            })
            self.assertEqual(rv['article'], {
                'id': 30, 'uid': 3
            })

        self.db.close()  # Reset session and its cache

        # Query get: relations
        with self.app.test_client() as c:
            rv = c.get('/article/30', json={
                'query': {
                    'project': ['uid'],
                    'join': ['user',]
                }
            })
            self.assertEqual(rv['article'], {
                'id': 30, 'uid': 3,
                'user': {
                    'id': 3, 'name': 'c',
                    'age': 16,
                    'tags': ['3', 'a', 'b', 'c'],
                }
            })

        self.db.close()  # Reset session and its cache

        # Query get: relations with filtering, projection and further joins
        with self.app.test_client() as c:
            rv = c.get('/article/30', json={
                'query': {
                    'project': ['id'],
                    'join': {
                        'user': {
                            'project': ['name'],
                            'filter': {'id': 3},
                            'join': ['comments'],
                        }
                    }
                }
            })
            self.assertEqual(rv['article'], {
                'id': 30,
                'user': {
                    'id': 3, 'name': 'c',
                    'comments': [{'id': 3, 'uid': 3, 'aid': 10, 'text': '10-c', }]
                }
            })

        self.db.close()  # Reset session and its cache

        # Query get: try banned relation
        with self.app.test_client() as c:
            self.assertRaises(AssertionError, c.get, '/article/30', json={
                'query': {
                    'join': ['comments'],
                }
            })

    def test_update(self):
        """ Test update() """

        # Update
        # `uid` should be copied over
        # JSON `data` should be merged
        with self.app.test_client() as c:
            rv = c.post('/article/10', json={
                'article': {
                    'id': 999, 'uid': 999, # 'ro': ignored
                    'data': {'?': ':)'}
                }
            })
            self.assertEqual(rv['article'], {
                'id': 10,  # ro
                'uid': 1,  # ro
                'title': '10',  # Unchanged
                'data': {'?': ':)', 'o': {'a': True}, 'rating': 5},  # merged
            })

    def test_delete(self):
        """ Test delete() """

        # Delete
        with self.app.test_client() as c:
            rv = c.delete('/article/10', json=None)
            self.assertEqual(rv['article'], {
                'id': 10, 'uid': 1,
                'title': '10',
                'data': {'o': {'a': True}, 'rating': 5},

                # FIXME: for some reason, deleted entity has ALL relationships loaded! Why? They shouldn't be here
                'user': {'age': 18, 'id': 1, 'name': 'a', 'tags': ['1', 'a']},
                'comments': [{'aid': None, 'id': 1, 'text': '10-a', 'uid': 1},{'aid': None, 'id': 2, 'text': '10-b', 'uid': 2},{'aid': None, 'id': 3, 'text': '10-c', 'uid': 3}],
            })

            self.db.close()

            self.assertRaises(NoResultFound, c.get, '/article/10')  # really removed

    def test_404(self):
        """ Try accessing entities that do not exist """



from collections import OrderedDict

from sqlalchemy import Integer, Float
from sqlalchemy.orm import load_only, defaultload, lazyload, contains_eager, aliased
from sqlalchemy.sql.expression import and_, or_, not_, cast
from sqlalchemy.sql import operators
from sqlalchemy.sql.functions import func

from sqlalchemy.dialects import postgresql as pg


class MongoProjection(object):
    """ MongoDB projection operator

        * { a: 1, b: 1 } - include only the given fields
        * { a: 0, b: 0 } - exlude the given fields
        * [ a, b, c ] - include only the given fields
    """

    def __init__(self, projection):
        """ Create a projection

            :type projection: None | Sequence | dict
            :raises AssertionError: invalid input
        """

        #: Inclusion mode or exclusion mode?
        self.inclusion_mode = False

        #: Normalized projection: { field_name: 0|1 }
        self.projection = {}

        # Empty projection
        if not projection:
            self.inclusion_mode = False
            self.projection = {}
            return

        # Array syntax
        if isinstance(projection, (list, tuple)):
            self.inclusion_mode = True
            self.projection = {k: 1 for k in projection}
            return

        # Object syntax
        assert isinstance(projection, dict), 'Projection must be one of: None, list, dict'
        assert sum(projection.values()) in [0, len(projection)], 'Dict projection values shall be all 0s or all 1s'

        self.projection = projection
        self.inclusion_mode = any(projection.values())

    @classmethod
    def columns(cls, bag, projection, inclusion_mode):
        """ Get the list of columns to be included

            :type bag: mongosql.bag.ModelPropertyBags
            :rtype: sqlalchemy.orm.Load
            :return: Options to include columns
            :raises AssertionError: unknown column name
        """
        # Check columns
        projection_keys = set(projection.keys())
        assert projection_keys <= bag.columns.names, 'Invalid column specified in projection'

        if inclusion_mode:
            return (bag.columns[name] for name in projection.keys())
        else:
            return (col for name, col in bag.columns.items() if name not in projection_keys)

    @classmethod
    def options(cls, bag, projection, inclusion_mode, as_relation):
        """ Get query options for the columns """
        return [as_relation.load_only(c) for c in cls.columns(bag, projection, inclusion_mode)]

    def __call__(self, model, as_relation):
        """ Build the statement

            :type model: MongoModel
            :param as_relation: Load interface to chain the loader options from
            :type as_relation: sqlalchemy.orm.Load
            :rtype: list[sqlalchemy.sql.schema.Column]
            :return: The list of columns to include
            :raises AssertionError: unknown column name
        """
        return self.options(model.model_bag, self.projection, self.inclusion_mode, as_relation)


class MongoSort(object):
    """ MongoDB sorting

        * OrderedDict({ a: +1, b: -1 })
        * [ 'a+', 'b-', 'c' ]  - array of strings '<column>[<+|->]'. default direction = +1
    """

    def __init__(self, sort_spec):
        """ Create the sorter

            :type sort_spec: string | Sequence | OrderedDict
        """

        #: Normalized sort: { field: +1 | -1 }
        self.sort = OrderedDict()

        # Empty
        if not sort_spec:
            sort_spec = []

        # List
        if isinstance(sort_spec, (list, tuple)):
            # Strings
            if all(isinstance(v, basestring) for v in sort_spec):
                sort_spec = OrderedDict([
                    [v[:-1], -1 if v[-1] == '-' else +1]
                    if v[-1] in {'+', '-'}
                    else [v, +1]
                    for v in sort_spec
                ])

        # OrderedDict
        if isinstance(sort_spec, OrderedDict):
            self.sort = sort_spec

            # Check directions
            assert all(dir in {-1, +1} for field, dir in self.sort.items()), '{} direction can be either +1 or -1'.format(type(self).__name__)
            return

        # Otherwise
        raise AssertionError('{} must be one of: None, list, OrderedDict'.format(type(self).__name__))

    @classmethod
    def columns(cls, bag, sort):
        """ Get the list of sorters for columns

            :type bag: mongosql.bag.ModelPropertyBags
            :rtype: list
            :return: The list of sort specifications

            :raises AssertionError: unknown column name
        """
        return [
            bag.columns[name].desc() if d == -1 else bag.columns[name]
            for name, d in sort.items()
        ]

    def __call__(self, model):
        """ Build the statement

            :type model: MongoModel
            :rtype: list
            :return: Sort columns

            :raises AssertionError: unknown column name
        """
        return self.columns(model.model_bag, self.sort)


class MongoGroup(MongoSort):
    """ MongoDB-style grouping

        See :cls:MongoSort
    """

    def __init__(self, group_spec):
        """ Create the grouper

            :type group_spec: string | Sequence | OrderedDict
        """
        super(MongoGroup, self).__init__(group_spec)

    def __call__(self, model):
        """ Build the statement

            :type model: MongoModel
            :rtype: list
            :return: The list of column groupers

            :raises AssertionError: unknown column name
        """
        return self.columns(model.model_bag, self.sort)


class MongoCriteria(object):
    """ MongoDB criteria

        Supports the following MongoDB operators:

        * { a: 1 }  - equality. For arrays: contains value.
        * { a: { $lt: 1 } }  - <
        * { a: { $lte: 1 } } - <=
        * { a: { $ne: 1 } } - <>. For arrays: does not contain value
        * { a: { $gte: 1 } } - >=
        * { a: { $gt: 1 } } - >
        * { a: { $in: [...] } } - any of. For arrays: has any from
        * { a: { $nin: [...] } } - none of. For arrays: has none from

        * { a: { $exists: true } } - is [not] NULL

        * { arr: { $all: [...] } } For arrays: contains all values
        * { arr: { $size: 0 } } For arrays: has a length of 0

        Supports the following boolean operators:

        * { $or: [ {..criteria..}, .. ] }  - any is true
        * { $and: [ {..criteria..}, .. ] } - all are true
        * { $nor: [ {..criteria..}, .. ] } - none is true
        * { $not: { ..criteria.. } } - negation
    """

    def __init__(self, criteria):
        """ Init a criteria

        :param criteria: Criteria
        :type criteria: None | dict
        """
        if not criteria:
            criteria = {}
        assert isinstance(criteria, dict), 'Criteria must be one of: None, dict'
        self.criteria = criteria

    # noinspection PyComparisonWithNone
    @classmethod
    def statement(cls, bag, criteria):
        """ Create a statement from criteria
        :type bag: mongosql.bag.ModelPropertyBags
        :rtype: sqlalchemy.sql.elements.BooleanClauseList
        """
        # Assuming a dict of { column: value } and { column: { $op: value } }

        conditions = []
        for col_name, criteria in criteria.items():
            # Boolean expressions?
            if col_name in {'$or', '$and', '$nor'}:
                assert isinstance(criteria, (list, tuple)), 'Criteria: {} argument must be a list'.format(col_name)
                if len(criteria) == 0:
                    continue  # skip empty

                criteria = map(lambda s: cls.statement(bag, s), criteria)  # now a list of expressions
                if col_name == '$or':
                    cc = or_(*criteria)
                elif col_name == '$and':
                    cc = and_(*criteria)
                elif col_name == '$nor':
                    cc = or_(*criteria)
                    conditions.append(~cc.self_group() if len(criteria) > 1 else ~cc)
                    continue
                else:
                    raise KeyError('Unknown operator '+col_name)

                conditions.append(cc.self_group() if len(criteria) > 1 else cc)
                continue
            elif col_name == '$not':
                assert isinstance(criteria, dict), 'Criteria: $not argument must be a dict'
                criteria = cls.statement(bag, criteria)
                conditions.append(not_(criteria))
                continue

            # Prepare
            col = bag.columns[col_name]
            col_array = bag.columns.is_column_array(col_name)
            col_json  = bag.columns.is_column_json(col_name)

            # Fake equality
            if not isinstance(criteria, dict):
                criteria = {'$eq': criteria}  # fake the missing equality operator for simplicity

            # Iterate over operators
            for op, value in criteria.items():
                value_array = isinstance(value, (list, tuple))

                # Coerce operand
                if col_array and value_array:
                    value = cast(pg.array(value), pg.ARRAY(col.type.item_type))
                if col_json:
                    coerce_type = col.type.coerce_compared_value('=', value)  # HACKY: use sqlalchemy type coercion
                    col = cast(col.astext, coerce_type)

                # Operators
                if op == '$eq':
                    if col_array:
                        if value_array:
                            conditions.append(col == value)  # Array equality
                        else:
                            conditions.append(col.any(value))  # ANY(array) = value, for scalar values
                    else:
                        conditions.append(col == value)  # array == value, for both scalars
                elif op == '$ne':
                    if col_array and not value_array:
                        if value_array:
                            conditions.append(col != value)  # Array inequality
                        else:
                            conditions.append(col.all(value, operators.ne))  # ALL(array) != value, for scalar values
                    else:
                        conditions.append(col != value)  # array != value, for both scalars
                elif op == '$lt':
                    conditions.append(col < value)
                elif op == '$lte':
                    conditions.append(col <= value)
                elif op == '$gte':
                    conditions.append(col >= value)
                elif op == '$gt':
                    conditions.append(col > value)
                elif op == '$in':
                    assert value_array, 'Criteria: $in argument must be a list'
                    if col_array:
                        conditions.append(col.overlap(value))  # field && ARRAY[values]
                    else:
                        conditions.append(col.in_(value))  # field IN(values)
                elif op == '$nin':
                    assert value_array, 'Criteria: $nin argument must be a list'
                    if col_array:
                        conditions.append(~ col.overlap(value))  # NOT( field && ARRAY[values] )
                    else:
                        conditions.append(col.notin_(value))  # field NOT IN(values)
                elif op == '$exists':
                    if value:
                        conditions.append(col != None)  # IS NOT NULL
                    else:
                        conditions.append(col == None)  # IS NULL
                elif op == '$all':
                    assert col_array, 'Criteria: $all can only be applied to an array column'
                    assert value_array, 'Criteria: $all argument must be a list'
                    conditions.append(col.contains(value))
                elif op == '$size':
                    assert col_array, 'Criteria: $all can only be applied to an array column'
                    if value == 0:
                        conditions.append(func.array_length(col, 1) == None)  # ARRAY_LENGTH(field, 1) IS NULL
                    else:
                        conditions.append(func.array_length(col, 1) == value)  # ARRAY_LENGTH(field, 1) == value
                else:
                    raise AssertionError('Criteria: unsupported operator "{}"'.format(op))
        if conditions:
            cc = and_(*conditions)
            return cc.self_group() if len(conditions) > 1 else cc
        else:
            return True

    def __call__(self, model):
        """ Build the statement

            :type model: MongoModel
            :return: SQL statement for filter()
            :rtype: sqlalchemy.sql.elements.BooleanClauseList
            :raises AssertionError: unknown column name
        """
        return self.statement(model.model_bag, self.criteria)


class _MongoJoinParams(object):
    def __init__(self, options, relationship=None, target_model=None, query=None):
        """ Values for joins
        :param options: Additional query options
        :type options: Sequence[sqlalchemy.orm.Load]
        :param relationship: Relationship
        :type relationship: sqlalchemy.orm.attributes.InstrumentedAttribute
        :param target_model: Target model
        :type target_model: sqlalchemy.ext.declarative.DeclarativeMeta
        :param query: Query dict for :meth:MongoQuery.query()
        :type query: dict
        """
        self.options = options
        self.relationship = relationship
        self.target_model = target_model
        self.query = query


class MongoJoin(object):
    """ Joining relations (eager load)

        - List of relation names
        - Dict: { relation-name: query-dict } for :meth:MongoQuery.query
    """

    def __init__(self, relnames):
        """ Create the joiner

        :param relnames: List of relation names to load eagerly
        :type relnames: Sequence[str]
        """

        if not relnames:
            self.rels = {}
        elif isinstance(relnames, (list, tuple)):
            self.rels = {relname: None for relname in relnames}
        elif isinstance(relnames, dict):
            self.rels = relnames
        else:
            raise AssertionError('Join must be one of: None, list, dict')

    @classmethod
    def options(cls, bag, rels, as_relation):
        """ Prepare relationships loader
        :type bag: mongosql.bag.ModelPropertyBags
        :type as_relation: sqlalchemy.orm.Load
        :returns: List of _MongoJoinParams
        :rtype: list[_MongoJoinParams]
        """
        relnames = set(rels.keys())
        assert relnames <= bag.relations.names, 'Invalid relation names: {}'.format(relnames - bag.relations.names)

        # Complex joins
        mjp_list = []
        for relname, query in rels.items():
            rel = bag.relations[relname]
            if query is None:
                # No query specified: don't load further relations
                as_relation.defaultload(rel).lazyload('*')
            else:
                # Query is present: prepare join information for further queries

                rel_a = aliased(rel)
                target_model = rel.property.mapper.class_

                mjp_list.append(_MongoJoinParams(
                    [as_relation.contains_eager(rel)],
                    rel,
                    target_model,
                    query
                ))

        # lazyload() on all other relations
        opts = [as_relation.lazyload(bag.relations[relname]) for relname in bag.relations.names if relname not in relnames]  # FIXME: apply lazyload() to all attributes initially, then override these. How do I do it?  http://stackoverflow.com/questions/25000473/
        mjp_list.append(_MongoJoinParams(opts))

        # Finish
        return mjp_list

    def __call__(self, model, as_relation):
        """ Build the statement

            :type model: MongoModel
            :param as_relation: Load interface to chain the loader options from
            :type as_relation: sqlalchemy.orm.Load
            :return: List of join params
            :rtype: list[_MongoJoinParams]
            :raises AssertionError: unknown column name
        """
        return self.options(model.model_bag, self.rels, as_relation)


class MongoAggregate(object):
    """ Aggregation statements

        { computed_field_name: aggregation-expression }

        Aggregation expressions:

            * column-name
            * { $min: operand }
            * { $max: operand }
            * { $avg: operand }
            * { $sum: operand }

        An operand can be:

            - Integer (for counting): { $sum: 1 }
            - Column name
            - Boolean expression (see :cls:MongoCriteria)
    """

    def __init__(self, agg_spec):
        """ Create aggregation
        :param agg_spec: Aggregation spec
        :type agg_spec: dict
        """
        if not agg_spec:
            agg_spec = {}
        assert isinstance(agg_spec, dict), 'Aggregate spec must be one of: None, dict'
        self.agg_spec = agg_spec

    @classmethod
    def selectables(cls, bag, agg_spec):
        """ Create a list of statements from spec

        :type bag: mongosql.bag.ModelPropertyBags
        :rtype: list[sqlalchemy.sql.elements.ColumnElement]
        """
        # TODO: calculation expressions for selection: http://docs.mongodb.org/manual/meta/aggregation-quick-reference/
        selectables = []
        for comp_field, comp_expression in agg_spec.items():
            # Column reference
            if isinstance(comp_expression, basestring):
                selectables.append(bag.columns[comp_expression].label(comp_field))
                continue

            # Computed expression
            assert isinstance(comp_expression, dict), 'Aggregate: Expression should be either a column name, or an object'
            assert len(comp_expression) == 1, 'Aggregate: expression can only contain a single operator'
            operator, expression = comp_expression.popitem()

            # Expression statement
            if isinstance(expression, int) and operator == '$sum':
                # Special case for count
                expression_stmt = expression
            elif isinstance(expression, basestring):
                # Column name
                expression_stmt = bag.columns[expression]
                # Json column?
                if bag.columns.is_column_json(expression):
                    # PostgreSQL always returns text values from it, and for aggregation we usually need numbers :)
                    expression_stmt = cast(expression_stmt, Float)
            elif isinstance(expression, dict):
                # Boolean expression
                expression_stmt = MongoCriteria.statement(bag, expression)
                # Need to cast it to int
                expression_stmt = cast(expression_stmt, Integer)
            else:
                raise AssertionError('Aggregate: expression should be either a column name, or an object')

            # Operator
            if operator == '$max':
                comp_stmt = func.max(expression_stmt)
            elif operator == '$min':
                comp_stmt = func.min(expression_stmt)
            elif operator == '$avg':
                comp_stmt = func.avg(expression_stmt)
            elif operator == '$sum':
                if isinstance(expression_stmt, int):
                    # Special case for count
                    comp_stmt = func.count()
                    if expression_stmt != 1:
                        comp_stmt *= expression_stmt
                else:
                    comp_stmt = func.sum(expression_stmt)
            else:
                raise AssertionError('Aggregate: unsupported operator "{}"'.format(operator))

            # Append
            selectables.append(comp_stmt.label(comp_field))

        return selectables

    def __call__(self, model):
        """ Build the statement

            :type model: MongoModel
            :return: List of selectables
            :rtype: list[sqlalchemy.sql.elements.ColumnElement]
            :raises AssertionError: wrong expression
        """
        return self.selectables(model.model_bag, self.agg_spec)

# TODO: update operations in MongoDB-style



import unittest
import re
from collections import OrderedDict

from sqlalchemy.orm import Query
from sqlalchemy.dialects import postgresql as pg

from . import models


def q2sql(q):
    """ Convert an SqlAlchemy query to string """
    # See: http://stackoverflow.com/a/4617623/134904
    # This intentionally does not escape values!
    stmt = q.statement
    dialect = pg.dialect()
    query = stmt.compile(dialect=dialect)
    return (query.string.encode(dialect.encoding) % query.params).decode(dialect.encoding)


class StatementsTest(unittest.TestCase):
    """ Test statements as strings """

    longMessage = True

    def test_projection(self):
        """ Test project() """
        m = models.User

        project = lambda projection: m.mongoquery(Query([m])).project(projection).end()

        def test_projection(projection, expected_columns):
            qs = q2sql(project(projection))
            rex = re.compile('u\.(\w+)[, ]')
            self.assertSetEqual(set(rex.findall(qs)), set(expected_columns), 'Expected {} in {}'.format(expected_columns, qs))

        # Empty values
        test_projection(None, ('id', 'name', 'tags', 'age'))
        test_projection([], ('id', 'name', 'tags', 'age'))
        test_projection({}, ('id', 'name', 'tags', 'age'))

        # Array syntax
        test_projection(['id'], ('id',))
        test_projection(['id', 'name'], ('id', 'name'))
        test_projection(['name'], ('id', 'name',))  # PK is always included :)
        self.assertRaises(AssertionError, project, ['id', 'lol'])

        # Object: inclusion
        test_projection({'id': 1}, ('id',))
        test_projection({'id': 1, 'name': 1}, ('id', 'name'))
        test_projection({'name': 1}, ('id', 'name',))

        # Object: exclusion
        test_projection({'id': 0}, ('id', 'name', 'tags', 'age'))
        test_projection({'id': 0, 'name': 0}, ('id', 'tags', 'age'))
        test_projection({'name': 0}, ('id', 'tags', 'age'))

        # Object: invalid
        self.assertRaises(AssertionError, project, {'id': 1, 'lol': 1})
        self.assertRaises(AssertionError, project, {'id': 0, 'lol': 0})
        self.assertRaises(AssertionError, project, {'id': 1, 'name': 0})

    def test_sort(self):
        """ Test sort() """
        m = models.User

        sort = lambda sort_spec: m.mongoquery(Query([m])).sort(sort_spec).end()

        def test_sort(sort_spec, expected_ends):
            qs = q2sql(sort(sort_spec))
            self.assertTrue(qs.endswith(expected_ends), '{!r} should end with {!r}'.format(qs, expected_ends))

        # Empty
        test_sort(None, u'FROM u')
        test_sort([], u'FROM u')
        test_sort(OrderedDict(), u'FROM u')

        # List
        test_sort(['id-', 'age-'], 'ORDER BY u.id DESC, u.age DESC')

        # Dict
        test_sort(OrderedDict([['id', -1], ['age', -1]]), 'ORDER BY u.id DESC, u.age DESC')

        # Fail
        self.assertRaises(AssertionError, test_sort, OrderedDict([['id', -2], ['age', -1]]), '')

    def test_group(self):
        """ Test group() """
        m = models.User

        group = lambda group_spec: m.mongoquery(Query([m])).group(group_spec).end()

        def test_group(group_spec, expected_ends):
            qs = q2sql(group(group_spec))
            self.assertTrue(qs.endswith(expected_ends), '{!r} should end with {!r}'.format(qs, expected_ends))

        # Empty
        test_group(None, u'FROM u')
        test_group([], u'FROM u')
        test_group(OrderedDict(), u'FROM u')

        # List
        test_group(['id-', 'age-'], 'GROUP BY u.id DESC, u.age DESC')

        # Dict
        test_group(OrderedDict([['id', -1], ['age', -1]]), 'GROUP BY u.id DESC, u.age DESC')

        # Fail
        self.assertRaises(AssertionError, test_group, OrderedDict([['id', -2], ['age', -1]]), '')

    def test_filter(self):
        """ Test filter() """
        m = models.User

        filter = lambda criteria: m.mongoquery(Query([m])).filter(criteria).end()

        def test_filter(criteria, expected):
            qs = q2sql(filter(criteria))
            self.assertEqual(qs.partition('\nWHERE ')[2], expected)

        # Empty
        test_filter(None, 'true')
        test_filter({}, 'true')

        # Wrong
        self.assertRaises(AssertionError, test_filter, [1, 2], '')

        # Equality, multiple
        test_filter({'id': 1, 'name': 'a'}, '(u.id = 1 AND u.name = a)')
        test_filter({'tags': 'a'}, 'a = ANY (u.tags)')
        test_filter({'tags': ['a', 'b', 'c']}, 'u.tags = CAST(ARRAY[a, b, c] AS VARCHAR[])')

        # $ne
        test_filter({'id': {'$ne': 1}}, 'u.id != 1')
        test_filter({'tags': {'$ne': 'a'}}, 'a != ALL (u.tags)')
        test_filter({'tags': {'$ne': ['a', 'b', 'c']}}, "u.tags != CAST(ARRAY[a, b, c] AS VARCHAR[])")

        # $lt, $lte, $gte, $gt
        test_filter({'id': {'$lt': 1}},  'u.id < 1')
        test_filter({'id': {'$lte': 1}}, 'u.id <= 1')
        test_filter({'id': {'$gte': 1}}, 'u.id >= 1')
        test_filter({'id': {'$gt': 1}},  'u.id > 1')

        # $in
        self.assertRaises(AssertionError, filter, {'tags': {'$in': 1}})
        test_filter({'name': {'$in': ['a', 'b', 'c']}}, 'u.name IN (a, b, c)')
        test_filter({'tags': {'$in': ['a', 'b', 'c']}}, 'u.tags && CAST(ARRAY[a, b, c] AS VARCHAR[])')

        # $nin
        self.assertRaises(AssertionError, filter, {'tags': {'$nin': 1}})
        test_filter({'name': {'$nin': ['a', 'b', 'c']}}, 'u.name NOT IN (a, b, c)')
        test_filter({'tags': {'$nin': ['a', 'b', 'c']}}, 'NOT u.tags && CAST(ARRAY[a, b, c] AS VARCHAR[])')

        # $exists
        test_filter({'name': {'$exists': 0}}, 'u.name IS NULL')
        test_filter({'name': {'$exists': 1}}, 'u.name IS NOT NULL')

        # $all
        self.assertRaises(AssertionError, filter, {'name': {'$all': ['a', 'b', 'c']}})
        self.assertRaises(AssertionError, filter, {'tags': {'$all': 1}})
        test_filter({'tags': {'$all': ['a', 'b', 'c']}}, "u.tags @> CAST(ARRAY[a, b, c] AS VARCHAR[])")

        # $size
        self.assertRaises(AssertionError, filter, {'name': {'$size': 0}})
        test_filter({'tags': {'$size': 0}}, "array_length(u.tags, 1) IS NULL")
        test_filter({'tags': {'$size': 1}}, "array_length(u.tags, 1) = 1")

        # $or
        self.assertRaises(AssertionError, filter, {'$or': {}})
        test_filter({'$or': [{'id': 1}, {'name': 'a'}]}, "(u.id = 1 OR u.name = a)")

        # $and
        self.assertRaises(AssertionError, filter, {'$and': {}})
        test_filter({'$and': [{'id': 1}, {'name': 'a'}]}, "(u.id = 1 AND u.name = a)")

        # $nor
        self.assertRaises(AssertionError, filter, {'$nor': {}})
        test_filter({'$nor': [{'id': 1}, {'name': 'a'}]}, "NOT (u.id = 1 OR u.name = a)")

        # $not
        self.assertRaises(AssertionError, filter, {'$not': []})
        test_filter({'$not': {'id': 1}}, "u.id != 1")

        # Braces
        self.assertRaises(AssertionError, filter, {'$or': {}})
        test_filter({'$or': [{'id': 1, 'name': 'a'}, {'name': 'b'}]}, "((u.id = 1 AND u.name = a) OR u.name = b)")

    def test_limit(self):
        """ Test limit() """
        m = models.User

        limit = lambda limit=None, skip=None: m.mongoquery(Query([m])).limit(limit, skip).end()

        def test_limit(lim, skip, expected_endswith):
            qs = q2sql(limit(lim, skip))
            self.assertTrue(qs.endswith(expected_endswith), '{!r} should end with {!r}'.format(qs, expected_endswith))

        # Skip
        test_limit(None, None, 'FROM u')
        test_limit(None, -1, 'FROM u')
        test_limit(None, 0, 'FROM u')
        test_limit(None, 1, 'LIMIT ALL OFFSET 1')
        test_limit(None, 9, 'LIMIT ALL OFFSET 9')

        # Limit
        test_limit(-1, None, 'FROM u')
        test_limit(0, None, 'FROM u')
        test_limit(1, None, 'LIMIT 1')
        test_limit(9, None, 'LIMIT 9')

        # Both
        test_limit(5, 10, 'LIMIT 5 OFFSET 10')

        # Twice
        q = limit(limit=10)
        q = m.mongoquery(q).limit(limit=15, skip=30).end()
        qs = q2sql(q)
        self.assertTrue(qs.endswith('LIMIT 15 OFFSET 30'), qs)

    def test_join(self):
        """ Test join() """
        m = models.User

        # Okay
        mq = m.mongoquery(Query([models.User]))
        mq = mq.join(('articles', 'comments'))
        q = mq.end()
        qs = q2sql(q)
        self.assertIn('FROM u', qs)
        self.assertIn('LEFT OUTER JOIN a', qs)
        self.assertIn('LEFT OUTER JOIN c', qs)

        # Unknown relation
        mq = m.mongoquery(Query([models.User]))
        self.assertRaises(AssertionError, mq.join, ('???'))

    def test_aggregate(self):
        """ Test aggregate() """
        m = models.User

        aggregate = lambda agg_spec: m.mongoquery(Query([m])).project(('id',)).aggregate(agg_spec).end()

        def test_aggregate(agg_spec, expected_starts):
            qs = q2sql(aggregate(agg_spec))
            self.assertTrue(qs.startswith(expected_starts), '{!r} should start with {!r}'.format(qs, expected_starts))

        # Empty
        test_aggregate(None, 'SELECT u.id AS u_id \nFROM')
        test_aggregate({},   'SELECT u.id AS u_id \nFROM')

        # $func(column)
        test_aggregate({ 'max_age': {'$max': 'age'} }, 'SELECT max(u.age) AS max_age \nFROM')
        test_aggregate({ 'min_age': {'$min': 'age'} }, 'SELECT min(u.age) AS min_age \nFROM')
        test_aggregate({ 'avg_age': {'$avg': 'age'} }, 'SELECT avg(u.age) AS avg_age \nFROM')
        test_aggregate({ 'sum_age': {'$sum': 'age'} }, 'SELECT sum(u.age) AS sum_age \nFROM')

        # $sum(1)
        test_aggregate({'count': {'$sum': 1}}, 'SELECT count(*) AS count')
        test_aggregate({'count': {'$sum': 10}}, 'SELECT count(*) * 10 AS count')

        # $sum(id==1)
        test_aggregate({'count': {'$sum': { 'id': 1 } }}, 'SELECT sum(CAST(u.id = 1 AS INTEGER)) AS count \nFROM')

        # age, $sum(1)
        q = OrderedDict()  # OrderedDict() to have predictable output
        q['age'] = 'age'  # column
        q['n'] = {'$sum': 1}
        test_aggregate(q, 'SELECT u.age AS age, count(*) AS n \nFROM')

        # $max(age), $sum(id=1 AND age >= 16)
        q = OrderedDict()  # OrderedDict() to have predictable output
        q['max_age'] = {'$max': 'age'}
        q['count'] = {'$sum': OrderedDict([('id', 1), ('age', {'$gte': 16})])}
        test_aggregate(q, 'SELECT max(u.age) AS max_age, sum(CAST((u.id = 1 AND u.age >= 16) AS INTEGER)) AS count \nFROM')

        # Unknown column
        self.assertRaises(AssertionError, test_aggregate, {'a': '???'}, '')
        self.assertRaises(AssertionError, test_aggregate, {'a': {'$max': '???'}}, '')
        self.assertRaises(AssertionError, test_aggregate, {'a': {'$sum': {'???': 1}}}, '')



from copy import copy
from sqlalchemy import Column
from sqlalchemy.orm import ColumnProperty, RelationshipProperty
from sqlalchemy.orm import Session, exc
from sqlalchemy.util import KeyedTuple
from sqlalchemy.orm.session import make_transient, make_transient_to_detached

from . import MongoModel, MongoQuery


def expunge_instance(instance):
    """ Expunge an instance from its session.

    Sometimes we keep the previous version of an instance.
    In order to avoid SqlAlchemy conflicts because of two instances existing at the same time,
    we need to remove the previous version from the session.

    :param instance: Instance
    :type instance: sqlalchemy.ext.declarative.DeclarativeMeta|None
    :rtype: Session
    """
    if not instance:
        return
    try:
        session = Session.object_session(instance)
        if session:
            session.expunge(instance)
        return session
    except exc.UnmappedInstanceError:
        pass


class CrudHelper(object):
    """ Crud helper functions """

    def __init__(self, model):
        """ Init CRUD helper

        :param model: The model to work with
        :type model: type
        """
        self.model = model
        self.mongomodel = MongoModel.get_for(self.model)

    def mquery(self, query, query_obj=None):
        """ Construct a MongoQuery for the model.

        If `query` is provided, it's used for initial filtering

        :param query: Query to start with
        :type query: sqlalchemy.orm.Query
        :param query_obj: Apply initial filtering with the Query Object
        :type query_obj: dict|None
        :rtype: mongosql.MongoQuery
        :raises AssertionError: unknown operations specified in query_obj
        """
        assert query_obj is None or isinstance(query_obj, dict), 'Query Object should be a dict or None'

        mq = MongoQuery(self.mongomodel, query)
        if query_obj:
            mq = mq.query(**query_obj)
        return mq

    def check_columns(self, names):
        """ Test if all column names are known

        :param names: Column names
        :type names: Iterable
        :return: Set of unknown names
        :rtype: set
        """
        model_colnames = self.mongomodel.model_bag.columns.names
        names = set(names)
        return names - model_colnames

    def nullify_empty_fields(self, entity):
        """ Walk through the entity dict and handle nullable fields:

        - If a field has a value of '', set it to None

        :param entity: Entity
        :type entity:
        :return: Altered entity
        :rtype: dict
        """
        for k in self.mongomodel.model_bag.nullable.keys():
            if k in entity and entity[k] == '':
                entity[k] = None
        return entity

    def create_model(self, entity):
        """ Create an instance from entity dict.

        This only allows to assign column properties and not relations.

        :param entity: Entity dict
        :type entity: dict
        :return: Created instance
        :rtype: sqlalchemy.ext.declarative.DeclarativeMeta
        :raises AssertionError: validation errors
        """
        assert isinstance(entity, dict), 'Create model: entity should be a dict'

        # Check columns
        unk_cols = self.check_columns(entity.keys())
        assert not unk_cols, 'Create model: unknown fields: {}'.format(unk_cols)

        # Create
        return self.model(**entity)

    def instance_for_update(self, prev_instance):
        """ Based on `prev_instance`, create a new instance that will track updates.

            To achieve that, we first create an instance based on `prev_instance`'s primary key, and then mark it as "Detached":
            then SqlAlchemy thinks it was persisted previously, and therefore tracks property updates on it.

            Also, `prev_instance` is expunge()d from its session, and a new instance takes its place.

            :param prev_instance: Origin.
            :type prev_instance: sqlalchemy.ext.declarative.DeclarativeMeta
            :rtype: sqlalchemy.ext.declarative.DeclarativeMeta
        """
        # Create from PK
        pk_values = {c: getattr(prev_instance, c) for c in self.mongomodel.model_bag.pk.names}
        new_instance = self.model(**pk_values)

        # Make 'detached'
        make_transient_to_detached(new_instance)

        # Expunge previous, replace with new
        ssn = expunge_instance(prev_instance).add(new_instance)
        return new_instance

    def update_model(self, entity, prev_instance):
        """ Update an instance from entity dict by merging the fields

        - Properties are copied over
        - JSON dicts are shallowly merged

        :param entity: Entity dict
        :type entity: dict
        :param prev_instance: Previous version
        :type prev_instance: sqlalchemy.ext.declarative.DeclarativeMeta
        :return: New instance, updated
        :rtype: sqlalchemy.ext.declarative.DeclarativeMeta
        :raises AssertionError: validation errors
        """
        assert isinstance(entity, dict), 'Update model: entity should be a dict'

        # Make a fresh instance
        # We need this magic to preserve `prev_instance` intact so `update_hook` can deal with it.
        # Also, this allows to use just Session.add() instead or Session.merge()
        new_instance = self.instance_for_update(prev_instance)

        # Check columns
        unk_cols = self.check_columns(entity.keys())
        assert not unk_cols, 'Update model: unknown fields: {}'.format(unk_cols)

        # Update
        for name, val in entity.items():
            if isinstance(val, dict) and self.mongomodel.model_bag.columns.is_column_json(name):
                # JSON column
                # NOTE: the field is very capricious to change management!
                p = dict(getattr(new_instance, name) or ())  # Defaults to empty dict
                for k, v in val.items():
                    p[k] = v  # Can't use update(): psycopg then raises 'TypeError: can't escape unicode to binary' o_O
                setattr(new_instance, name, p)  # so SQLalchemy knows the field is updated
            else:
                # Other columns
                setattr(new_instance, name, val)

        # Finish
        return new_instance


class StrictCrudHelper(CrudHelper):
    """ Crud helper with limitations

        - Read-only fields can not be set
        - Only allowed relationships can be loaded
        - Default Query Object is used
        - Limits the maximum number of items that can be retrieved when listing
    """

    def __init__(self, model, ro_fields=(), allow_relations=(), query_defaults=None, maxitems=None):
        """ Init Strict CRUD helper

        :param model: The model to work with
        :type model: sqlalchemy.ext.declarative.DeclarativeMeta
        :param ro_fields: List of read-only properties.

            Also can be a callable which decides on the read-only properties at runtime.

        :type ro_fields: Iterable[str|sqlalchemy.Column|sqlalchemy.orm.properties.ColumnProperty]|Callable
        :param allow_relations: List of relations allowed to join to.
            Specify relation names or relationship properties.
            To allow joining to second-level relations, use dot-notation.
        :type allow_relations: Iterable[str|sqlalchemy.orm.relationships.RelationshipProperty]
        :param query_defaults: Default Query Object used when nothing is specified
        :type query_defaults: dict|None
        :param maxitems: Hard limit on the number of entities that can be loaded (max value for QueryObject['limit'])
        :type maxitems: int|None
        """
        super(StrictCrudHelper, self).__init__(model)

        self._ro_fields = ro_fields if callable(ro_fields) else set(c.key if isinstance(c, (Column, ColumnProperty)) else c for c in ro_fields)
        self._allowed_relations = set(c.key if isinstance(c, RelationshipProperty) else c for c in allow_relations)
        self._query_defaults = query_defaults or {}
        self._maxitems = maxitems or None

        assert callable(self._ro_fields) or all(map(lambda v: isinstance(v, str), self._ro_fields)), 'Some values in `ro_fields` were not converted to string'
        assert all(map(lambda v: isinstance(v, str), self._allowed_relations)), 'Some values in `allowed_relations` were not converted to string'
        assert isinstance(self._query_defaults, dict), '`query_defaults` was not a dict'
        assert self._maxitems is None or isinstance(self._maxitems, int), '`maxitems` must be an integer'

    @property
    def ro_fields(self):
        """ Get the set of read-only properties

        :rtype: set
        """
        return set(self._ro_fields()) if callable(self._ro_fields) else self._ro_fields

    @classmethod
    def _check_relations(cls, allowed_relations, qo, _prefix=''):
        """ Test Query Object joins against `allowed_relations`, supporting dot-notation

        :param allowed_relations: Set of allowed relations
        :type allowed_relations: set
        :param qo: Query Object
        :type qo: dict | None
        :returns: Banned relationships
        :rtype: set
        """
        if not isinstance(qo, dict) or 'join' not in qo:
            return set()
        joinspec = qo['join']

        relnames = {_prefix + name for name in joinspec}
        disallowed_relations = relnames - allowed_relations

        # Deeper
        if isinstance(joinspec, dict):
            for relname, qo in joinspec.items():
                disallowed_relations |= cls._check_relations(allowed_relations, qo, _prefix=relname + '.')

        # Finish
        return disallowed_relations

    def mquery(self, query, query_obj=None):
        assert query_obj is None or isinstance(query_obj, dict), 'Query Object should be a dict or None'

        # Query defaults
        if self._query_defaults:
            query_obj = dict(self._query_defaults.items() + (query_obj.items() if query_obj else []))

        # Max items
        if self._maxitems:
            query_obj = query_obj or {}
            if not (query_obj.get('count', 0) or query_obj.get('aggregate', 0)):  # no limits in count() and aggregate() modes
                query_obj['limit'] = min(self._maxitems, query_obj.get('limit', self._maxitems))

        # Allowed relations
        disallowed_relations = self._check_relations(self._allowed_relations, query_obj)
        assert not disallowed_relations, 'Joining to these relations is not allowed: {}'.format(disallowed_relations)

        # Finish
        return super(StrictCrudHelper, self).mquery(query, query_obj)

    def create_model(self, entity):
        assert isinstance(entity, dict), 'Create model: entity should be a dict'

        # Remove ro fields
        for k in set(entity.keys()) & self.ro_fields:
            entity.pop(k)

        # Super
        return super(StrictCrudHelper, self).create_model(entity)

    def update_model(self, entity, prev_instance):
        assert isinstance(entity, dict), 'Update model: entity should be a dict'

        # Remove ro fields
        for k in set(entity.keys()) & self.ro_fields:
            entity.pop(k)

        # Super
        return super(StrictCrudHelper, self).update_model(entity, prev_instance)


class CrudViewMixin(object):
    """ Base class for CRUD implementations """

    #: Set the CRUD helper object
    crudhelper = None

    @classmethod
    def _getCrudHelper(cls):
        """ Get the CRUD helper assigned for this class

        :rtype: mongosql.CrudHelper
        """
        assert isinstance(cls.crudhelper, CrudHelper), '{}: {} should be set to an instance of {}'.format(cls, 'crudhelper', CrudHelper)
        return cls.crudhelper

    def _query(self):
        """ Get a Query object to be used for queries

        :rtype: sqlalchemy.orm.Query
        """
        raise NotImplementedError('query() method not defined on {}'.format(type(self)))

    def _mquery(self, query_obj=None, *filter, **filter_by):
        """ Get a MongoQuery with initial filtering applied

        :param query_obj: Query Object
        :type query_obj: dict|None
        :param filter: Additional filter() criteria
        :param filter_by: Additional filter_by() criteria
        :rtype: mongosql.MongoQuery
        """
        return self._getCrudHelper().mquery(
            self._query().filter(*filter).filter_by(**filter_by),
            query_obj
        )

    def _get_one(self, query_obj, *filter, **filter_by):
        """ Utility method that fetches a single entity.

        You will probably want to override it with custom error handling

        :param query_obj: Query Object
        :param filter: Additional filter() criteria
        :param filter_by: Additional filter_by() criteria
        :rtype: sqlalchemy.ext.declarative.DeclarativeMeta
        :raises sqlalchemy.orm.exc.NoResultFound: Nothing found
        :raises sqlalchemy.orm.exc.MultipleResultsFound: Multiple found
        :raises AssertionError: validation errors
        """
        return self._mquery(query_obj, *filter, **filter_by).end().one()

    def _save_hook(self, new, prev=None):
        """ Hook into create(), update() methods.

        This allows to make some changes to the instance before it's actually saved.

        :param new: New version
        :type new: sqlalchemy.ext.declarative.DeclarativeMeta
        :param prev: Previously persisted version (only available for 'update')
        :type prev: sqlalchemy.ext.declarative.DeclarativeMeta|None
        """
        pass

    def _method_list(self, query_obj=None, *filter, **filter_by):
        """ Fetch the list of entitites

        :param query_obj: Query Object
        :param filter: Additional filter() criteria
        :param filter_by: Additional filter_by() criteria
        :rtype: list
        :raises AssertionError: validation errors
        """
        res = self._mquery(query_obj, *filter, **filter_by).end().all()

        # Count?
        if query_obj and query_obj.get('count', 0):
            return res[0][0]  # Scalar count query

        # Convert KeyedTuples to dicts (when aggregating)
        if query_obj and 'aggregate' in query_obj:
            return [dict(zip(row.keys(), row)) for row in res]

        return res

    def _method_create(self, entity):
        """ Create a new entity

        :param entity: Entity dict
        :type entity: dict
        :return: The created instance (to be saved)
        :rtype: sqlalchemy.ext.declarative.DeclarativeMeta
        :raises AssertionError: validation errors
        """
        instance = self._getCrudHelper().create_model(entity)
        self._save_hook(instance)
        return instance

    def _method_get(self, query_obj=None, *filter, **filter_by):
        """ Fetch a single entity

        :param query_obj: Query Object
        :param filter: Additional filter() criteria
        :param filter_by: Additional filter_by() criteria
        :rtype: sqlalchemy.ext.declarative.DeclarativeMeta
        :raises sqlalchemy.orm.exc.NoResultFound: Nothing found
        :raises sqlalchemy.orm.exc.MultipleResultsFound: Multiple found
        :raises AssertionError: validation errors
        """
        return self._get_one(query_obj, *filter, **filter_by)

    def _method_update(self, entity, *filter, **filter_by):
        """ Update an existing entity by merging the fields

        :param entity: Entity dict
        :type entity: dict
        :param filter: Criteria to find the previous entity
        :param filter_by: Criteria to find the previous entity
        :return: The updated instance (to be saved)
        :rtype: sqlalchemy.ext.declarative.DeclarativeMeta
        :raises sqlalchemy.orm.exc.NoResultFound: Nothing found
        :raises sqlalchemy.orm.exc.MultipleResultsFound: Multiple found
        :raises AssertionError: validation errors
        """
        prev_instance = self._get_one(None, *filter, **filter_by)
        new_instance = self._getCrudHelper().update_model(entity, prev_instance)

        self._save_hook(new_instance, prev_instance)
        return new_instance

    def _method_delete(self, *filter, **filter_by):
        """ Delete an existing entity

        Loads the entity prior to deletion.

        :param filter: Criteria to find the previous entity
        :param filter_by: Criteria to find the previous entity
        :return: The instance to be deleted
        :rtype: sqlalchemy.ext.declarative.DeclarativeMeta
        :raises sqlalchemy.orm.exc.NoResultFound: Nothing found
        :raises sqlalchemy.orm.exc.MultipleResultsFound: Multiple found
        :raises AssertionError: validation errors
        """
        return self._get_one(None, *filter, **filter_by)



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, Integer
from sqlalchemy.orm import relationship, backref
from sqlalchemy.sql.schema import ForeignKey

from sqlalchemy.dialects import postgresql as pg

from mongosql import MongoSqlBase
from flask.ext.jsontools import JsonSerializableBase


Base = declarative_base(cls=(MongoSqlBase, JsonSerializableBase))


class User(Base):
    __tablename__ = 'u'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    tags = Column(pg.ARRAY(String))  # ARRAY field
    age = Column(Integer)


class Article(Base):
    __tablename__ = 'a'

    id = Column(Integer, primary_key=True)
    uid = Column(Integer, ForeignKey(User.id))
    title = Column(String)
    data = Column(pg.JSON)  # JSON field

    user = relationship(User, lazy='joined', backref=backref('articles', lazy='joined',))


class Comment(Base):
    __tablename__ = 'c'

    id = Column(Integer, primary_key=True)

    aid = Column(Integer, ForeignKey(Article.id))
    uid = Column(Integer, ForeignKey(User.id))

    text = Column(String)

    article = relationship(Article, lazy='joined', backref=backref("comments", lazy='joined',))
    user = relationship(User, lazy='joined', backref=backref("comments", lazy='joined',))




def init_database():
    """ Init DB
    :rtype: (sqlalchemy.engine.Engine, sqlalchemy.orm.Session)
    """
    engine = create_engine('postgresql://postgres:postgres@localhost/test_mongosql', convert_unicode=True)
    Session = sessionmaker(autocommit=True, autoflush=True, bind=engine)
    return engine, Session

def create_all(engine):
    """ Create all tables """
    Base.metadata.create_all(bind=engine)

def drop_all(engine):
    """ Drop all tables """
    Base.metadata.drop_all(bind=engine)

def content_samples():
    """ Generate content samples """
    return [
        User(id=1, name='a', age=18, tags=['1', 'a']),
        User(id=2, name='b', age=18, tags=['2', 'a', 'b']),
        User(id=3, name='c', age=16, tags=['3', 'a', 'b', 'c']),

        Article(id=10, uid=1, title='10', data={'rating': 5  , 'o': {'a': True}}),
        Article(id=11, uid=1, title='11', data={'rating': 5.5, 'o': {'a': True}}),
        Article(id=12, uid=1, title='12', data={'rating': 6  , 'o': {'a': False}}),
        Article(id=20, uid=2, title='20', data={'rating': 4.5, 'o': {'a': False}}),
        Article(id=21, uid=2, title='21', data={'rating': 4  , 'o': {'z': True}}),
        Article(id=30, uid=3, title='30', data={               'o': {'z': False}}),

        Comment(aid=10, uid=1, text='10-a'),
        Comment(aid=10, uid=2, text='10-b'),
        Comment(aid=10, uid=3, text='10-c'),
        Comment(aid=11, uid=1, text='11-a'),
        Comment(aid=11, uid=2, text='11-b'),
        Comment(aid=12, uid=1, text='12-a'),
        Comment(aid=20, uid=1, text='20-a-ONE'),
        Comment(aid=20, uid=1, text='20-a-TWO'),
        Comment(aid=21, uid=1, text='21-a'),
    ]

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)

    engine, Session = init_database()
    ssn = Session()

    from sqlalchemy import inspect, func
    from sqlalchemy.orm import noload, load_only, defaultload, lazyload, aliased, contains_eager, contains_alias

    ssn.query(User).filter_by(id=999).delete()

    u1 = User(id=999, name=999)
    u2 = User(id=999, name=999) ; ssn.add(u2)
    u3 = ssn.query(User).filter_by(id=1).one()
    u4 = ssn.query(User).options(load_only(User.name)).filter_by(id=2).one() ; u4.tags = [1,2,3]

    #ssn.begin()
    #ssn.commit()


    s1, s2, s3, s4 = map(inspect, (u1, u2, u3, u4))

    from IPython import embed ; embed()



from sqlalchemy.orm import Session

from .model import MongoModel
from .query import MongoQuery


class MongoSqlBase(object):
    """ Mixin for SqlAlchemy models

        Provides methods for accessing :cls:MongoModel and :cls:MongoQuery
    """

    __mongomodel = None

    @classmethod
    def mongomodel(cls):
        """ Get MongoModel object
        :rtype: mongosql.MongoModel
        """
        if cls.__mongomodel is None:
            cls.__mongomodel = MongoModel(cls)
        return cls.__mongomodel

    @classmethod
    def mongoquery(cls, query, **kwargs):
        """ Build a MongoQuery
        :param query: Query to start with, or a session object to initiate the query with
        :type query: sqlalchemy.orm.Query|sqlalchemy.orm.Session
        :rtype: mongosql.MongoQuery
        """
        if isinstance(query, Session):
            query = query.query(cls)
        return MongoQuery(cls.mongomodel(), query, **kwargs)



import unittest
from sqlalchemy import inspect

from . import models


row2dict = lambda row: dict(zip(row.keys(), row))  # zip into a dict


class QueryTest(unittest.TestCase):
    """ Test MongoQuery """

    def setUp(self):
        # Connect, create tables
        engine, Session = models.init_database()
        models.drop_all(engine)
        models.create_all(engine)

        # Fill DB
        ssn = Session()
        ssn.begin()
        ssn.add_all(models.content_samples())
        ssn.commit()

        # Session
        self.Session = Session
        self.engine = engine
        self.db = Session()

        # Logging
        import logging
        logging.basicConfig(level=logging.DEBUG)
        logging.getLogger('sqlalchemy.engine').setLevel(logging.INFO)

    def tearDown(self):
        self.db.close()  # Need to close the session: otherwise, drop_all() hangs forever
        models.drop_all(self.engine)

    def test_projection(self):
        """ Test project() """
        ssn = self.db

        # Test: load only 2 props
        user = models.User.mongoquery(ssn).project(['id', 'name']).end().first()
        self.assertEqual(inspect(user).unloaded, {'age', 'tags', 'articles', 'comments'})

        # Test: load without 2 props
        user = models.User.mongoquery(ssn).project({'age': 0, 'tags': 0}).end().first()
        self.assertEqual(inspect(user).unloaded, {'age', 'tags', 'articles', 'comments'})

    def test_sort(self):
        """ Test sort() """
        ssn = self.db

        # Test: sort(age+, id-)
        users = models.User.mongoquery(ssn).sort(['age+', 'id+']).end().all()
        self.assertEqual([3, 1, 2], [u.id for u in users])

    def test_filter(self):
        """ Test filter() """
        ssn = self.db

        # Test: filter(age=16)
        users = models.User.mongoquery(ssn).filter({'age': 16}).end().all()
        self.assertEqual([3], [u.id for u in users])

    def test_join(self):
        """ Test join() """
        ssn = self.db

        # Test: no join(), relationships are unloaded
        user = models.User.mongoquery(ssn).end().first()
        self.assertEqual(inspect(user).unloaded, {'articles', 'comments'})

        # Test:    join(), relationships are   loaded
        user = models.User.mongoquery(ssn).join(['articles']).end().first()
        self.assertEqual(inspect(user).unloaded, {'comments'})

    def test_join_query(self):
        """ Test join(dict) """
        ssn = self.Session()

        # Test: join() with comments as dict
        user = models.User.mongoquery(ssn)\
            .filter({'id': 1})\
            .join({
                'comments': None
            })\
            .end().one()
        self.assertEqual(user.id, 1)
        self.assertEqual(inspect(user).unloaded, {'articles'})

        ssn.close() # need to reset the session: it caches entities and gives bad results

        # Test: join() with filtered articles
        user = models.User.mongoquery(ssn) \
            .filter({'id': 1}) \
            .join({
                'articles': {
                    'project': ['id', 'title'],
                    'filter': {'id': 10},
                    'limit': 1
                }
            }) \
            .end().one()
        self.assertEqual(user.id, 1)
        self.assertEqual(inspect(user).unloaded, {'comments'})
        self.assertEqual([10], [a.id for a in user.articles])  # Only one article! :)
        self.assertEqual(inspect(user.articles[0]).unloaded, {'user', 'comments',  'uid', 'data'})  # No relationships loaded, and projection worked

    def test_count(self):
        """ Test count() """
        ssn = self.db

        # Test: count()
        n = models.User.mongoquery(ssn).count().end().scalar()
        self.assertEqual(3, n)

    def test_aggregate(self):
        """ Test aggregate() """
        ssn = self.db

        # Test: aggregate()
        q = {
            'max_age': {'$max': 'age'},
            'adults': {'$sum': {'age': {'$gte': 18}}},
        }
        row = models.User.mongoquery(ssn).aggregate(q).end().one()
        ':type row: sqlalchemy.util.KeyedTuple'
        self.assertEqual(row2dict(row), {'max_age': 18, 'adults': 2})

        # Test: aggregate { $sum: 1 }
        row = models.User.mongoquery(ssn).aggregate({ 'n': {'$sum': 1} }).end().one()
        self.assertEqual(row.n, 3)

        # Test: aggregate { $sum: 10 }, with filtering
        row = models.User.mongoquery(ssn).filter({'id': 1}).aggregate({'n': {'$sum': 10}}).end().one()
        self.assertEqual(row.n, 10)

        # Test: aggregate() & group()
        q = {
            'age': 'age',
            'n': {'$sum': 1},
        }
        rows = models.User.mongoquery(ssn).aggregate(q).group(['age']).sort(['age-']).end().all()
        self.assertEqual(map(row2dict, rows), [ {'age': 18, 'n': 2}, {'age': 16, 'n': 1} ])

    def test_json(self):
        """ Test operations on a JSON column """
        ssn = self.db

        # Filter: >=
        articles = models.Article.mongoquery(ssn).filter({ 'data.rating': {'$gte': 5.5} }).end().all()
        self.assertEqual({11, 12}, {a.id for a in articles})

        # Filter: == True
        articles = models.Article.mongoquery(ssn).filter({'data.o.a': True}).end().all()
        self.assertEqual({10, 11}, {a.id for a in articles})

        # Filter: is None
        articles = models.Article.mongoquery(ssn).filter({'data.o.a': None}).end().all()
        self.assertEqual({21, 30}, {a.id for a in articles})

        # Filter: wrong type, but still works
        articles = models.Article.mongoquery(ssn).filter({'data.rating': '5.5'}).end().all()
        self.assertEqual({11}, {a.id for a in articles})

        # Sort
        articles = models.Article.mongoquery(ssn).sort(['data.rating-']).end().all()
        self.assertEqual([None, 6, 5.5, 5, 4.5, 4], [a.data.get('rating', None) for a in articles])

        # Aggregate
        q = {
            'high': {'$sum': { 'data.rating': {'$gte': 5.0} }},
            'max_rating': {'$max': 'data.rating'},
            'a_is_none': {'$sum': { 'data.o.a': None } },
        }
        row = models.Article.mongoquery(ssn).aggregate(q).end().one()
        self.assertEqual(row2dict(row), {'high': 3, 'max_rating': 6, 'a_is_none': 2})

        # Aggregate & Group



from mongosql import CrudViewMixin, StrictCrudHelper

from . import models
from flask import request, g
from flask.ext.jsontools import jsonapi, RestfulView


class ArticlesView(RestfulView, CrudViewMixin):
    """ Full-featured CRUD view """

    # We're strict, yeah
    crudhelper = StrictCrudHelper(models.Article,
        ro_fields=('id', 'uid',),
        allow_relations=('user', 'user.comments'),
        query_defaults={
            'sort': ['id-'],
        },
        maxitems=2,
    )

    # RestfulView needs that for routing
    primary_key = ('id',)
    decorators = (jsonapi,)

    # DRY: store it once
    entity_name = 'article'

    @property
    def _qo(self):
        """ Get Query Object from request

        :rtype: dict | None
        """
        return (request.get_json() or {}).get('query', None)

    def _db(self):
        """ Get database Session

        :rtype: sqlalchemy.orm.Session
        """
        return g.db

    def _query(self):  # Implemented
        return self._db().query(self.crudhelper.model)

    #region Collection methods

    def list(self):
        return { self.entity_name+'s': self._method_list(self._qo) }

    def create(self):
        instance = self._method_create(request.get_json()[self.entity_name])
        instance.uid = 3  # Manually set ro field value

        ssn = self._db()
        ssn.add(instance)
        ssn.commit()

        return {self.entity_name: instance}

    #endregion

    #region Single entity methods

    def get(self, id):
        return { self.entity_name: self._method_get(self._qo, id=id) }

    def update(self, id):
        instance = self._method_update(request.get_json()[self.entity_name], id=id)

        ssn = self._db()
        ssn.add(instance)
        ssn.commit()

        return {self.entity_name: instance}

    def delete(self, id):
        instance = self._method_delete(id=id)

        ssn = self._db()
        ssn.delete(instance)
        ssn.commit()

        return {self.entity_name: instance}

    #endregion



from sqlalchemy.orm import Query, Load, defaultload
from sqlalchemy.orm.query import QueryContext
from sqlalchemy.sql import func

from .model import MongoModel


class MongoQuery(object):
    """ MongoDB-style queries """

    @classmethod
    def get_for(cls, model, *args, **kwargs):
        """ Get MongoQuery for a model.

        Attempts to use `mongoquery` property of the model

        :param model: Model
        :type model: mongosql.MongoSqlBase|sqlalchemy.ext.declarative.DeclarativeMeta
        :rtype: MongoQuery
        """
        try:
            return model.mongoquery(*args, **kwargs)
        except AttributeError:
            return cls(MongoModel.get_for(model), *args, **kwargs)

    def __init__(self, model, query, _as_relation=None):
        """ Init a MongoDB-style query
        :param model: MongoModel
        :type model: mongosql.MongoModel
        :param query: Query to work with
        :type query: sqlalchemy.orm.Query
        :param _as_relation: Parent relationship.
            Internal argument used when working with deeper relations:
            is used as initial path for defaultload(_as_relation).lazyload(...).
        :type _as_relation: sqlalchemy.orm.relationships.RelationshipProperty
        """
        assert isinstance(model, MongoModel)
        assert isinstance(query, Query)

        self._model = model
        self._query = query
        self._as_relation = defaultload(_as_relation) if _as_relation else Load(self._model.model)

        self._no_joindefaults = False

    def aggregate(self, agg_spec):
        """ Select aggregated results """
        a = self._model.aggregate(agg_spec)
        if a:
            self._query = self._query.with_entities(*a)
            self._no_joindefaults = True  # no relationships should be loaded

            # When no model criteria is specified, like COUNT(*), SqlAlchemy won't set the FROM clause
            # Thus, we need to explicitly set the `FROM` clause in these cases
            if self._query.whereclause is None:
                self._query = self._query.select_from(self._model.model)

        return self

    def project(self, projection):
        """ Apply a projection to the query """
        p = self._model.project(projection, as_relation=self._as_relation)
        if self._model.model.__name__ == 'User':
            assert 1
        self._query = self._query.options(p)
        return self

    def sort(self, sort_spec):
        """ Apply sorting to the query """
        s = self._model.sort(sort_spec)
        self._query = self._query.order_by(*s)
        return self

    def group(self, group_spec):
        """ Apply grouping to the query """
        g = self._model.group(group_spec)
        self._query = self._query.group_by(*g)
        return self

    def filter(self, criteria):
        """ Add criteria to the query """
        c = self._model.filter(criteria)
        self._query = self._query.filter(c)
        return self

    def limit(self, limit=None, skip=None):
        """ Slice results """
        limit, skip = self._model.limit(limit, skip)
        if skip:
            self._query = self._query.offset(skip)
        if limit:
            self._query = self._query.limit(limit)
        return self

    def join(self, relnames):
        """ Eagerly load relations """
        for mjp in self._model.join(relnames, as_relation=self._as_relation):
            # Complex joins
            if mjp.query:
                self._query = self.get_for(
                        mjp.target_model,
                        self._query.join(mjp.relationship),
                        _as_relation=mjp.relationship
                    )\
                    .query(**mjp.query)\
                    .end()

            # Options
            self._query = self._query.options(*mjp.options)

        self._query = self._query.with_labels()
        self._no_joindefaults = True
        return self

    def count(self):
        """ Count rows instead """
        self._query = self._query.from_self(func.count(1))
        self._no_joindefaults = True  # no relationships should be loaded
        return self

    def query(self, project=None, sort=None, group=None, filter=None, skip=None, limit=None, join=None, aggregate=None, count=False, **__unk):
        """ Build a query
        :param project: Projection spec
        :param sort: Sorting spec
        :param group: Grouping spec
        :param filter: Filter criteria
        :param skip: Skip rows
        :param limit: Limit rows
        :param join: Eagerly load relations
        :param aggregate: Select aggregated results
        :param count: True to count rows instead
        :raises AssertionError: unknown Query Object operations provided (extra keys)
        :rtype: MongoQuery
        """
        assert not __unk, 'Unknown Query Object operations: {}'.format(__unk.keys())

        q = self
        if join:            q = q.join(join)
        if project:         q = q.project(project)
        if aggregate:       q = q.aggregate(aggregate)
        if filter:          q = q.filter(filter)
        if sort:            q = q.sort(sort)
        if group:           q = q.group(group)
        if skip or limit:   q = q.limit(limit, skip)
        return q.count() if count else q

    def end(self):
        """ Get the Query object
        :rtype: sqlalchemy.orm.Query
        """
        if not self._no_joindefaults:
            self.join(())  # have to join with an empty list explicitly so all relations get noload()
        return self._query



from sqlalchemy import inspect
from sqlalchemy.dialects import postgresql as pg


class _PropertiesBag(object):
    # region Protected

    @staticmethod
    def _is_column_array(col):
        """ Is the column an ARRAY column?

        :type col: sqlalchemy.sql.schema.Column
        :rtype: bool
        """
        return isinstance(col.type, pg.ARRAY)

    @staticmethod
    def _is_column_json(col):
        """ Is the column a JSON column?

        :type col: sqlalchemy.sql.schema.Column
        :rtype: bool
        """
        return isinstance(col.type, pg.JSON)

    @staticmethod
    def _dot_notation(name):
        """ Split a property name that's using dot-notation

        :type name: str
        :rtype: str, list[str]
        """
        path = name.split('.')
        return path[0], path[1:]

    #endregion

    def __contains__(self, name):
        """ Test if the property is in the bag
        :param name: Property name
        :type name: str
        :rtype: bool
        """
        raise NotImplementedError

    def __getitem__(self, name):
        """ Get the property by name
        :param name: Property name
        :type name: str
        :rtype: sqlalchemy.orm.interfaces.MapperProperty
        """
        raise NotImplementedError


class ColumnsBag(_PropertiesBag):
    """ Columns bag """

    def __init__(self, columns):
        """ Init columns
        :param columns: Model columns
        :type columns: dict[sqlalchemy.orm.properties.ColumnProperty]
        """
        self._columns = columns
        self._column_names = set(self._columns.keys())
        self._array_columns = {name: col for name, col in self._columns.items() if self._is_column_array(col)}
        self._json_columns =  {name: col for name, col in self._columns.items() if self._is_column_json(col)}

    def is_column_array(self, name):
        """ Is the column an ARRAY column
        :type name: str
        :rtype: bool
        """
        column_name = self._dot_notation(name)[0]
        return column_name in self._array_columns

    def is_column_json(self, name):
        """ Is the column a JSON column
        :type name: str
        :rtype: bool
        """
        column_name = self._dot_notation(name)[0]
        return column_name in self._json_columns

    @property
    def names(self):
        """ Get the set of column names
        :rtype: set[str]
        """
        return self._column_names

    def items(self):
        """ Get columns
        :rtype: dict[sqlalchemy.orm.properties.ColumnProperty]
        """
        return self._columns.items()

    def __getitem__(self, column_name):
        try:
            return self._columns[column_name]
        except KeyError:
            raise AssertionError('Unknown column: `{}`'.format(column_name))


class PrimaryKeyBag(ColumnsBag):
    """ Primary Key Bag """


class DotColumnsBag(ColumnsBag):
    """ Columns bag with additional capabilities:

        - For JSON fields: field.prop.prop -- dot-notation access to sub-properties
    """

    def __getitem__(self, name):
        column_name, path = self._dot_notation(name)
        col = super(DotColumnsBag, self).__getitem__(column_name)
        # JSON path
        if path and self.is_column_json(column_name):
            col = col[path].astext
        return col


class RelationshipsBag(_PropertiesBag):
    """ Relationships bag with additional capabilities """

    def __init__(self, relationships):
        """ Init relationships
        :param relationships: Model relationships
        :type relationships: dict[sqlalchemy.orm.relationships.RelationshipProperty]
        """
        self._rels = relationships
        self._rel_names = set(self._rels.keys())

    @property
    def names(self):
        """ Get the set of relation names
        :rtype: set[str]
        """
        return self._rel_names

    def items(self):
        """ Get relationships
        :rtype: dict[sqlalchemy.orm.relationships.RelationshipProperty]
        """
        return self._rels.items()

    def __contains__(self, name):
        return name in self._rels

    def __getitem__(self, name):
        try:
            return self._rels[name]
        except KeyError:
            raise AssertionError('Unknown relationship: `{}`'.format(name))


class ModelPropertyBags(object):
    """ Model property bags """

    def __init__(self, model):
        """ Init bags

        :param model: Model
        :type model: sqlalchemy.ext.declarative.DeclarativeMeta
        """
        ins = inspect(model)
        self.model = model

        #: Column properties
        self.columns   =    DotColumnsBag({name:   getattr(model, name)   for name, c in ins.column_attrs .items()})

        #: Relationship properties
        self.relations = RelationshipsBag({name:   getattr(model, name)   for name, c in ins.relationships.items()})

        #: Primary key properties
        self.pk        =    PrimaryKeyBag({c.name: self.columns[c.name]   for       c in ins.primary_key})

        #: Nullable properties
        self.nullable  =       ColumnsBag({name: c for name, c in self.columns.items() if c.nullable})



from sqlalchemy import inspect

from .statements import MongoProjection, MongoSort, MongoGroup, MongoCriteria, MongoJoin, MongoAggregate
from .bag import ModelPropertyBags


class MongoModel(object):
    """ Sqlalchemy Model wrapper that generates query chunks """

    @classmethod
    def get_for(cls, model):
        """ Get MongoModel for a model.

        Attempts to use `mongomodel` property of the model

        :param model: Model
        :type model: mongosql.MongoSqlBase|sqlalchemy.ext.declarative.DeclarativeMeta
        :rtype: MongoModel
        """
        try:
            return model.mongomodel()
        except AttributeError:
            model.mongomodel = MongoModel(model)
            return model.mongomodel

    def __init__(self, model):
        """ Create MongoSql from a model

            :type model: sqlalchemy.ext.declarative.declarative_base
            :param model: The model to build queries for
        """
        #: The model we're working with
        self.__model = model

        #: Property bags
        self.__bag = ModelPropertyBags(model)

    @property
    def model(self):
        """ Get model
        :rtype: sqlalchemy.ext.declarative.DeclarativeMeta
        """
        return self.__model

    @property
    def model_bag(self):
        """ Get model columns bag
        :rtype: mongosql.bag.ModelPropertyBags
        """
        return self.__bag

    #region Wrappers

    def project(self, projection, as_relation):
        """ Build projection for a Query

        :type projection: None | dict | Iterable
        :param projection: Projection spec
        :param as_relation: Load interface to chain the loader options from
        :type as_relation: sqlalchemy.orm.Load
        :returns: Query options to load specific columns
            Usage:
                p = MongoModel(User).project(['login', 'email'])
                query.options(p)
        :rtype: sqlalchemy.orm.Load
        :raises AssertionError: invalid input
        :raises AssertionError: unknown column name
        """
        return MongoProjection(projection)(self, as_relation)

    def sort(self, sort_spec):
        """ Build sorting for a Query

            :type sort_spec: None | OrderedDict | Iterable
            :param sort_spec: Sort spec
            :returns: List of columns, with ordering modifiers applied.
                Usage:
                    s = MongoModel(User).sort(['state+'])
                    query.order_by(*s)
            :rtype: sqlalchemy.orm.query.Query
            :raises AssertionError: invalid input
            :raises AssertionError: unknown column name
        """
        return MongoSort(sort_spec)(self)

    def group(self, group_spec):
        """ Build grouping for a Query

            See :meth:sort()

            :returns: List of columns, with grouping modifiers applied.
                Usage:
                    s = MongoModel(User).group(['state'])
                    query.group_by(*s)
            :rtype: sqlalchemy.orm.query.Query
            :raises AssertionError: invalid input
            :raises AssertionError: unknown column name
        """
        return MongoGroup(group_spec)(self)

    def filter(self, criteria):
        """ Build filtering condition for a Query

            :type criteria: None | dict
            :param criteria: The criteria to filter with
            :rtype: sqlalchemy.sql.elements.BooleanClauseList
            :returns: Filtering conditions.
                Usage:
                    c = MongoModel(User).filter({ 'id': 1 })
                    query.filter(c)
            :raises AssertionError: invalid input
            :raises AssertionError: unknown column name
        """
        return MongoCriteria(criteria)(self)

    def skip(self, skip=None):
        """ Build the rows offset value

        :type skip: int | None
        :param skip: The number of rows to skip
        :returns: Offset value.
            Usage:
                skip = MongoModel(User).skip(10)
                query.offset(skip)
        :rtype: int | None
        :raises AssertionError: invalid input
        """
        assert skip is None or isinstance(skip, int), 'Skip must be one of: None, int'
        return None if skip is None or skip <= 0 else skip

    def limit(self, limit=None, skip=None):
        """ Build the rows limit & offset

            :type limit: int | None
            :param limit: Row count limitation
            :type skip: int | None
            :param skip: The number of rows to skip
            :returns: (limit, offset)
                Usage:
                    skip, limit = MongoModel(User).limit(20, 10)
                    query.limit(limit).offset(skip)
            :rtype: (int|None, int|None)
            :raises AssertionError: invalid input
        """
        assert limit is None or isinstance(limit, int), 'Limit must be one of: None, int'
        return (
            None if limit is None or limit <= 0 else limit,
            self.skip(skip)
        )

    def join(self, relnames, as_relation):
        """ Build eager loader for the relations

            :type relnames: None | Iterable[str] | dict
            :param relnames: List of relations to load eagerly
            :param as_relation: Load interface to chain the loader options from
            :type as_relation: sqlalchemy.orm.Load
            :returns: Join params list
                Usage: don't ask :)
            :returns: (list[query-options], list[(join, query-dict)])
            :rtype: list[mongosql.statements._MongoJoinParams]
            :raises AssertionError: invalid input
        """
        return MongoJoin(relnames)(self, as_relation)

    def aggregate(self, agg_spec):
        """ Select aggregated results

        :param agg_spec: Aggregation spec
        :type agg_spec: None | dict
        :return: List of selectables.
                Usage:
                    a = MongoModel(User).aggregate({ oldest: { $max: 'age' } })
                    query.add_columns(*a)
        :rtype: list[sqlalchemy.sql.elements.ColumnElement]
        :raises AssertionError: invalid input
        """
        return MongoAggregate(agg_spec)(self)

    #endregion
